#include "runner/runner.hpp"
#include "bin/test_expecter.hpp"

#include <boost/program_options.hpp>
namespace po = boost::program_options;
#include <boost/exception/diagnostic_information.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/filesystem.hpp>
namespace fs = boost::filesystem;

#include <iostream>
#include <fstream>
#include <cstdlib>

#ifdef IS_LINUX
#include <unistd.h>
#endif


// Root Leaf data files (these are generated files, in the build directory)
#include "runner/unit_test.leaf.hpp"

shared_ptr<test_expecter> create_expecter( runner & run ) {
	//one of the files contains the test conditions
	std::string xcode;
	if( run.input_files.size() == 1 ) {
		xcode = util::load_file( run.input_files[0] );
	} else {
		for( auto & file : run.input_files ) {
			//only the full mode can have multiple files now, so this check is specific to that
			if( boost::ends_with( file, "main.leaf" ) ) {
				xcode = util::load_file( file );
			}
		}
	}
	
	auto exp = make_shared<test_expecter>(xcode);
	run.expect = exp;
	STATE_CHECK( exp->any_expect(), "no test conditions found" );
	return exp;
}

//there appears to be nothing in Boost that does this, and it handles `pine` quite differently than `pine/`
void force_directory( std::string & path ) {
	if( !boost::ends_with( path, "/" ) ) {
		path += "/";
	}
}

std::string get_leaf_path(char* arg0) {
	//figure out our directory first, before anything can interfere
	std::string our_dir_name;
	bool have = false;
	
#ifdef IS_LINUX
	ssize_t bufSize = 2000;
	char buf[bufSize];
	ssize_t res = readlink( "/proc/self/exe", buf, bufSize );
	if( res > 0 && res < bufSize)
	{
		our_dir_name = fs::path(buf).parent_path().generic_string();
		have = true;
	}
#endif

	if( !have) {
		auto our_path = fs::canonical(fs::system_complete(arg0));
		our_dir_name = our_path.parent_path().generic_string();
	}
	
	auto leaf_path = our_dir_name;
	char * leaf_path_c = std::getenv( "LEAF_PATH" );
	if( leaf_path_c ) {
		leaf_path = leaf_path_c;
	}
	force_directory(leaf_path);
	
	return leaf_path;
}

int main( int argc, char ** argv ) {
	
	try {
		//call first to avoid directory changing issues (if any)
		auto leaf_path = get_leaf_path(argv[0]);
		
		runner run;
		
		std::string expr, obj_file, cli_module;
		bool exe_test = false, unit_test = false, skip_omit = false, skip_ignore = false;
		po::options_description desc( "Options" );
		desc.add_options()
			( "help", "show help" )
			( "skip-omit", po::value( &skip_omit )->zero_tokens(), "skip checks marked as omit" )
			( "skip-ignore", po::value( &skip_ignore )->zero_tokens(), "skip checks marked as ignore" )
			( "unit-test", po::value( &unit_test )->zero_tokens(), "the file is a Leaf unit test, enable test functions" )
			( "exe-test", po::value( &exe_test )->zero_tokens(), "variant of unit-test which captures stdout" )
			( "obj-file", po::value( &obj_file ), "produce object file (provide basename)" )
			( "input-file", po::value( &run.input_files ), "leaf input files" )
			( "shared", po::value( &obj_file ), "generate a shared library" )
			( "exe", po::value( &obj_file ), "generate an executable" )
			( "library", po::value( &run.libraries ), "libraries to load prior to execution" )
			( "module", po::value( &cli_module ), "generate module in standard location with path name" )
			;
		po::positional_options_description p;
		p.add( "input-file", -1 );

		po::options_description dopts( "Leaf Debugging" );
		dopts.add_options()
			( "dump-parse", po::value( &run.dump_parse )->zero_tokens(), "dump the parse tree" )
			( "dump-parse-flat", po::value( &run.dump_parse_flat )->zero_tokens(), "use a flat parse tree dump" )
			( "dump-base", po::value( &run.dump_base )->zero_tokens(), "dump base content as well (must specify what to dump still)" )
			( "dump-pre-type", po::value( &run.dump_pre_type )->zero_tokens(), "dump the Leaf AST prior to typing" )
			( "dump-post-type", po::value( &run.dump_post_type )->zero_tokens(), "dump the Leaf AST post typing" )
			( "dump-all-types", po::value( &run.dump_all_types )->zero_tokens(), "dump all leaf AST tree types" )
			( "dump-ir", po::value( &run.dump_ir )->zero_tokens(), "dump the Leaf-IR" )
			( "dump-ir-types", po::value( &run.dump_ir_types )->zero_tokens(), "show all Leaf-IR types" )
			( "dump-llvm", po::value( &run.dump_llvm )->implicit_value("*"), "dump matching LLVM-IR symbols" )
			( "list-functions", po::value( &run.list_functions )->zero_tokens(), "list all function definitions" )
			;
			
		po::options_description opts( "Optimizations" );
		opts.add_options()
			( "llvm-opt", po::value( &run.gen_opts.llvm_opt )->zero_tokens(), "run LLVM optimizers" )
			( "reduce-branch", po::value( &run.gen_opts.reduce_branch )->implicit_value(true), "Use reduced branch flow" )
			( "reduce-flow", po::value( &run.gen_opts.reduce_flow )->implicit_value(true), "Use reduced blocks in flow" )
			;
			
		po::options_description all_opts;
		all_opts.add( desc ).add( opts ).add( dopts );
		
		po::variables_map vm;
		po::store( po::command_line_parser( argc, argv ).options( all_opts ).positional(p).run(), vm );
		po::notify( vm );

		if( vm.count( "help" ) ) {
			std::cerr << all_opts;
			return 0;
		}
		
		bool is_shared = vm.count("shared");
		
		//convenience to create a library in the standard location
		if( !cli_module.empty() ) {
			force_directory( cli_module );
			auto name = fs::path(cli_module).parent_path().filename().generic_string();
			is_shared = true;
			
			//only add location is not specified
			if( run.input_files.size() == 0 ) {
				run.input_files.push_back( cli_module );
			}
			
			if( obj_file.empty() ) {
				obj_file = leaf_path + std::string("libs/") + name;
			}
		}
		
		if( run.input_files.empty() ) {
			std::cerr << "requires input file" << std::endl;
			return 1;
		}
		
		//split obj_name
		std::string mod_name;
		if( obj_file.size() ) {
			fs::path pn(obj_file);
			fs::create_directories( pn.parent_path() );
			mod_name = pn.filename().string();
		}

		//name is in obj-file
		runner::gen_what_t gw = runner::gw_object;
		if( is_shared ) {
			gw = runner::gw_shared;
			run.gen_opts.pic = true;
			run.module_name = mod_name;
		} else if( vm.count( "exe" ) ) {
			gw = runner::gw_exe;
		}

		run.mloader = make_shared<leaf::module_loader>( leaf_path );
		
		if( exe_test ) {
			run.capture_stdout = true;
		}
	
		run.init_module();
		if( unit_test ) {
			run.add_base_block( run.create_string_source("unit_test.leaf", dataUnitTestLeaf ));
		}
		
		//prepare expectations
		shared_ptr<test_expecter> expect;
		if( unit_test || exe_test ) {
			expect = create_expecter(run);
			if( skip_omit && expect->omit ) {
				return 2;
			}
			if( skip_ignore && expect->ignore ) {
				return 3;
			}
		}

		auto module = run.build_module();
		if( !module ) {
			//happens only with expected failed compilation
			return 0;
		}

		if( obj_file.size() ) {
			run.emit_object( module, obj_file, gw );
			return 0;
		} else {
			return run.execute( module );	
		}
		
	} catch( recoverable_error & re ) {
		std::cerr << boost::current_exception_diagnostic_information();
		auto loc = boost::get_error_info<util::source_location::error>( re );
		if( loc )	{
			std::cout << "Source: " << loc->name()
				<< ":" << loc->line() << "," <<  loc->column()
				<< std::endl;
		}
		return 1;
	} catch(...) {
		std::cerr << boost::current_exception_diagnostic_information();
		return 1;
	}
}
